
from __future__ import print_function
import argparse
import time
import datetime
import json
import os
from os import listdir
from os.path import isfile, join
import numpy as np
import tensorflow as tf
from tflearn.data_utils import image_preloader
import matplotlib.pyplot as plt
from tropic_util import *

import keras
from keras import backend as K
from keras.preprocessing import image
from keras.preprocessing.image import ImageDataGenerator
#from keras.applications.resnet import ResNet152
from keras.applications.resnet50 import ResNet50
from keras.applications.inception_v3 import InceptionV3
#from keras.applications.resnet import ResNet50, ResNet152
from keras.layers import Dense, Conv2D, BatchNormalization, Activation
from keras.layers import GlobalAveragePooling2D, AveragePooling2D, Input, Flatten
from keras.layers.core import Dense, Dropout
from keras.regularizers import l2
from keras.models import Model
from keras.models import load_model
from keras.callbacks import ModelCheckpoint, LearningRateScheduler
from keras.callbacks import ReduceLROnPlateau
from keras.optimizers import Adam, SGD


np.set_printoptions(threshold=np.nan)

def model_resnetcount(weight):
    num_classes = 1
    base_model = ResNet50(weights=weight, include_top=False,input_tensor=Input(shape=(imw,imw,num_channels)))
        
    x = base_model.output
    #x = Flatten()(x)
    x = GlobalAveragePooling2D()(x)
    x = Dense(1024, activation='relu')(x)
    predictions = Dense(num_classes, activation='linear')(x)
    model = Model(inputs=base_model.input, outputs=predictions)
    
    model.compile(optimizer=SGD(lr=0.0001, momentum=0.9), loss='mean_absolute_error', metrics=['mae'])  #metrics=['mse','mae','accuracy']
    return model
    
def model_inceptioncount(weight):
    num_classes = 1
    base_model = InceptionV3(weights=weight, include_top=False,input_tensor=Input(shape=(imw,imw,num_channels)))
    x = base_model.output
    x = GlobalAveragePooling2D()(x)
    x = Dense(1024, activation='relu')(x)
    predictions = Dense(num_classes, activation='linear')(x)
    model = Model(inputs=base_model.input, outputs=predictions)
    model.compile(optimizer=SGD(lr=0.0001, momentum=0.9), loss='mean_absolute_error', metrics=['mae'])
    return model
        
    
def model_resnet(weight):
    
    base_model = ResNet50(weights=weight, include_top=False,input_tensor=Input(shape=(imw,imw,num_channels)))
        
    x = base_model.output
    x = Flatten()(x)
    predictions = Dense(num_classes, activation='linear')(x)
    model = Model(inputs=base_model.input, outputs=predictions)

    model.compile(optimizer=SGD(lr=0.0001, momentum=0.9), loss='mean_absolute_error', metrics=['mae'])  #metrics=['mse','mae','accuracy']
    return model
    
def model_inception(weight, datagen):
    
    base_model = InceptionV3(weights=weight, include_top=False,input_tensor=Input(shape=(imw,imw,num_channels)))
    x = base_model.output
    x = GlobalAveragePooling2D()(x)
    x = Dense(1024, activation='relu')(x)
    predictions = Dense(num_classes, activation='linear')(x)
    model = Model(inputs=base_model.input, outputs=predictions)
    
    '''
    print('\n\n________ layers name ___________\n')
    for i, layer in enumerate(base_model.layers):
        print(i, layer.name)
            
    model.compile(optimizer='adam', loss='mean_squared_error', metrics=['mse'])
    model.summary()
    model.fit_generator(datagen.flow(x_train, y_train, batch_size=batch_size), validation_data=(x_test, y_test), epochs=10, verbose=1, workers=4, shuffle=True)

    for layer in model.layers[:249]:
        layer.trainable = False
    for layer in model.layers[249:]:
        layer.trainable = True
    '''
    model.compile(optimizer=SGD(lr=0.0001, momentum=0.9), loss='mean_absolute_error', metrics=['mae'])
    return model
    

    
def save_history_graph(history):
    
    cdt = datetime.datetime.now()
    finish_time = ('%d-%d-%d_%d:%d'%(cdt.year,cdt.month,cdt.day,cdt.hour,cdt.minute))
    history_logfile = '%s/%s_%s_%s_%s_%s_%s.json'%(history_dir, dataset_name, exp_name, cnn, training_type, freeze, finish_time)
    print('_______________', history_logfile)
    with open(history_logfile,'w') as f:
        json.dump(history.history, f)                

    # Plot training & validation loss values
    image_lossfile = '%s/%s_%s_%s_%s_%s.png'%(graph_dir, dataset_name, exp_name, cnn, training_type, finish_time)
    ymin = 0
    ymax = 20
    axes = plt.gca()
    axes.set_ylim([ymin,ymax])
    plt.plot(history.history['loss'])
    plt.plot(history.history['val_loss'])
    plt.title('MAE loss : %s_%s_%s_%s'%(dataset_name, cnn, training_type, freeze))
    plt.ylabel('Loss')
    plt.xlabel('Epoch')
    plt.legend(['Train', 'Validation'], loc='upper left')
    plt.savefig(image_lossfile)
    #plt.show()
        
def train_data():
    start_time = time.time()
    
    input_shape = (imw, imw, num_channels)
        
    if not is_train:
        #evaluate_model(savedmodel_dir,None,x_test,y_test, count_only)
        model = None
    else:  #train model
        
        save_dir = os.path.join(os.getcwd(), savedmodel_dir)
        model_name = '%s_%s_%s_%s_%s_%s_imw%s_.{epoch:04d}.h5' % (dataset_name, exp_name, cnn, training_type, freeze, exp_name,imw)
        if not os.path.isdir(save_dir):
            os.makedirs(save_dir)
        filepath = os.path.join(save_dir, model_name)
            
        print('Using real-time data augmentation.')
        datagen = ImageDataGenerator(
            featurewise_center=False,
            samplewise_center=False,
            featurewise_std_normalization=False,
            samplewise_std_normalization=False,
            zca_whitening=False,
            rotation_range=0,
            width_shift_range=0.1,
            height_shift_range=0.1,
            horizontal_flip=True,
            vertical_flip=False)
        datagen.fit(x_train)
        
        checkpoint = ModelCheckpoint(filepath=filepath, monitor='val_loss', verbose=1)  
        callbacks = [checkpoint]
        
        if is_finetune:
            weight = 'imagenet'
        else:
            weight = None
            
        if cnn=='resnet':
            model = model_resnet(weight)
        elif cnn =='inception':
            model = model_inception(weight, datagen)
        elif cnn =='resnet_count':
            model = model_resnetcount(weight)
        elif cnn =='inception_count':
            model = model_inceptioncount(weight)
               

        history = model.fit_generator(datagen.flow(x_train, y_train, batch_size=batch_size),
                        validation_data=(x_test, y_test),
                        epochs=epochs, verbose=1, workers=4, shuffle=True, callbacks=callbacks)
        
        save_history_graph(history)       
         
        
    ##### end else
    ### get all loss and save to log file
                            
    end_time = time.time()
    (mse, rmse, mae,mae_count, mae_missed_classify, n_diff) = evaluate_model(savedmodel_dir,model,x_test,y_test, count_only)
    
    total_time = (int)(end_time-start_time)/60.0  

    write_to_logfile(logfile_report, dataset_name,exp_name,epochs, mse, rmse, mae , mae_count, mae_missed_classify, n_diff, total_time)
        
        
def print_working_detail():
    print('_____________working detail____________')
    print('%s, %s ,%s      '%(cnn, training_type, freeze))
    print("dataset:     ", dataset_name)
    print("exp:         ", exp_name)
    print("imw:         ", imw)
    print("batch_size:  ", batch_size)
    print('logfile_report:', logfile_report)
    print('savedmodel_dir:', savedmodel_dir)    
    print('_______________________________________')     
    
if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="TrainingModel")
    parser.add_argument('--dataset_name',   default ='dataset_augment', help='dataset name')
    parser.add_argument('--exp_name',       default ='exp5', help='exp name')    
    parser.add_argument('--imw',            default=224, type=int, help='Image size')
    parser.add_argument('--epochs',         default = 1000, type=int)
    parser.add_argument('--batch_size',     default = 64, type = int)
    parser.add_argument('--cnn',            default = 'inception', help='resnet, inception')
    parser.add_argument('--is_finetune',    type=str2bool, default=True)
    parser.add_argument('--is_freeze',      type=str2bool, default=False)    
    parser.add_argument('--is_train',       type=str2bool, default=True)
    parser.add_argument('--count_only',     type=str2bool, default=False, help='True if doesnt care about class')


    args = parser.parse_args()
    
    #assign arguments to global variables

    dataset_name    = args.dataset_name
    exp_name        = args.exp_name
    imw             = args.imw
    epochs          = args.epochs
    batch_size      = args.batch_size
    cnn             = args.cnn
    is_finetune     = args.is_finetune
    is_freeze       = args.is_freeze
    is_train        = args.is_train
    count_only      = args.count_only
    
    if is_finetune:
        training_type = 'finetune'
    else:
        training_type = 'scratch'
    
    if is_freeze:
        freeze = 'freeze'
    else:
        freeze = 'unfreeze'

    subtract_pixel_mean = True
    data_augmentation = True
    num_channels    = 3        
    num_classes = 5    
    
    if count_only:
        if cnn=='resnet':
            cnn = 'resnet_count'
        else:
            cnn = 'inception_count'
            
        logfile_report = 'countonly_resultlog/%s_%s_%s_%s.log'%(dataset_name, cnn, training_type, freeze)
        savedmodel_dir = 'countonly_savedmodel/%s_%s_%s_%s_%s'%(dataset_name, exp_name, cnn, training_type, freeze) 
        history_dir    = 'countonly_historylog'
        graph_dir      = 'countonly_graph'
    else:        
        logfile_report = 'mae_resultlog/%s_%s_%s_%s.log'%(dataset_name, cnn, training_type, freeze)
        savedmodel_dir = 'mae_savedmodel/%s_%s_%s_%s_%s'%(dataset_name, exp_name, cnn, training_type, freeze)
        history_dir    = 'mae_historylog'
        graph_dir      = 'mae_graph'
        #logfile_report    = 'regression_resultlog/%s_%s_%s_%s.log'%(dataset_name, cnn, training_type, freeze)
        #savedmodel_dir = 'regression_savedmodel/%s_%s_%s_%s_%s'%(dataset_name, exp_name, cnn, training_type, freeze)
        #history_dir    = 'regression_historylog'
        #graph_dir      = 'regression_graph'

    print_working_detail()
    
    x_train,y_train, x_test, y_test  = load_dataset(dataset_name, exp_name, imw, is_train, count_only)
    
    train_data()
